import java.util.*;
import java.io.*;

/**************************************************************
 * This program takes bowtie-aligned paired reads and runs EM to
 * learn edge weights on a new-format PSG.
 **************************************************************/

public class AlignPaired {
	
    public static SpliceGraph s;
	
	public static HashMap<String,ArrayList<Integer>> first_ids;
	public static HashMap<String,ArrayList<Integer>> second_ids;
	public static ArrayList<String> matched_keys;

    public static String[][] reads;
	public static ArrayList[] align;
	public static ArrayList[] confidence;
	public static ArrayList[] lengths;
	public static double[] c1;
	public static double[] c2;
	public static int fasta_readlen;
	
	public static ArrayList<Integer[]> align1;
	public static ArrayList<Integer[]> align2;
    
    public static void main(String[] args) {
		if (args.length != 4 && args.length != 7) {
			System.out.println("usage: AlignPaired <psg> <bowtie.1.align> <bowtie.2.align> <FASTA readlen> [<mean> <stdev> <beta>]");
			System.exit(0);
		}
		
		s = new SpliceGraph(args[0]);
		
		String filename1 = args[1];
		String filename2 = args[2];
		fasta_readlen = Integer.parseInt(args[3]);
		//System.err.println("Processing Reads...");
		process_reads(filename1,filename2);
		//System.err.println("Reads processed.");
		
		try {
			s.setMean(Integer.parseInt(args[4]));
			s.setStdev(Double.parseDouble(args[5]));
			s.setBeta(Double.parseDouble(args[6]));
		} catch (ArrayIndexOutOfBoundsException e) {
			// do nothing, this is okay
		}
		catch (NumberFormatException nfe) {
			System.out.print("Error: Mean should be given as an integer, ");
			System.out.println("Stdev should be given as a double.");
			System.exit(0);
		}
		
		// checking the edge case where we don't have any aligned reads
		if (reads.length == 0) {
			double ll = 0.0;
			//System.out.println("\tNumber of null reads: 0");
			System.out.println("\tNumber of reads: 0");
			s.emOutput(ll,-1);
			
			System.out.println();
			System.out.println("Current log likelihood: "+ll);
			System.out.println("Converged after 0 iterations");
			System.out.println("Log likelihood diff: 0");
			System.out.println("Max edge difference: 0.0");
			System.out.println("L1 norm: 0.0");
			System.out.println("L2 norm: 0.0");
			return;
		}
		
		// so if we're here, then we definitely have reads
		System.out.println("\tNumber of reads: "+reads.length);
		
		double[] initial = s.getEdgeWeights();
		s.resetEdgeWeights();
		int[][] length = Utility.IntArray(lengths);
		
		s.learnEM(reads,align,confidence,1,length);
		
		double[] a = s.getEdgeWeights();
		double result = 0;
		for (int i=0; i<initial.length; i++) {
			result += Math.abs(a[i]-initial[i]);
		}
		System.out.println("L1 norm: "+result);
		result = 0;
		for (int i=0; i<initial.length; i++) {
			result += Math.pow(a[i]-initial[i],2);
		}
		result = Math.sqrt(result);
		System.out.println("L2 norm: "+result);
    }
	
	// reads information from a bowtie alignment file
    private static void process_reads(String filename1, String filename2) {
		// initialize the temporary variables
		ArrayList<Integer> start1 = new ArrayList<Integer>();
		ArrayList<Integer> start2 = new ArrayList<Integer>();
		ArrayList<String> read1 = new ArrayList<String>();
		ArrayList<String> read2 = new ArrayList<String>();
		ArrayList<String> alig1 = new ArrayList<String>();
		ArrayList<String> alig2 = new ArrayList<String>();
		ArrayList<String> which1 = new ArrayList<String>();
		ArrayList<String> which2 = new ArrayList<String>();
		ArrayList<Integer> length1 = new ArrayList<Integer>();
		ArrayList<Integer> length2 = new ArrayList<Integer>();
		
		ArrayList<String> mismatch = new ArrayList<String>();
		ArrayList<String> quality = new ArrayList<String>();
		
		// initialize the global variables
		reads = null;
		align = null;
		confidence = null;
		c1 = null;
		c2 = null;

		// retrieve records from alignment file
		ArrayList<AlignmentRecord> first_records = Utility.read_alignment_file(filename1);
		ArrayList<AlignmentRecord> second_records = Utility.read_alignment_file(filename2);
		
		for (int i=0; i<first_records.size(); i++) {
			AlignmentRecord r = first_records.get(i);
			which1.add(r.id);
			alig1.add(r.alignment);
			start1.add(r.start_position);
			read1.add(r.read);
			length1.add(r.read_length);
			quality.add(r.quality);
			mismatch.add(r.mismatch);
		}

		c1 = Utility.compute_quality(quality,mismatch);
		first_ids = Utility.make_id_map(which1);
		
		mismatch = new ArrayList<String>();
		quality = new ArrayList<String>();
		
		for (int i=0; i<second_records.size(); i++) {
			AlignmentRecord r = second_records.get(i);
			which2.add(r.id);
			alig2.add(r.alignment);
			start2.add(r.start_position);
			read2.add(r.read);
			length2.add(r.read_length);
			quality.add(r.quality);
			mismatch.add(r.mismatch);
		}
		
		c2 = Utility.compute_quality(quality,mismatch);
		second_ids = Utility.make_id_map(which2);
		
		reads = getReads(read1,read2);
		
		if (reads.length > 0) {
			process_align(alig1,alig2,start1,start2,length1,length2);
		}
    }
	
	// retrieves the reads from the read arraylist
	private static String[][] getReads(ArrayList<String> reads_1, ArrayList<String> reads_2) {
		String[][] result = new String[count_ids()][2];
		
		int index = 0;
		Object[] id_1 = first_ids.keySet().toArray();
		for (int i=0; i<id_1.length; i++) {
			String read_id_1 = (String) id_1[i];
			String read_id_2 = read_id_1.substring(0,read_id_1.length()-1) + "2";

			if (second_ids.containsKey(read_id_2)) {
				result[index][0] = reads_1.get(first_ids.get(read_id_1).get(0).intValue());
				result[index][1] = reads_2.get(second_ids.get(read_id_2).get(0).intValue());
				index++;
			}
		}
		return result;
    }
    
    // determines the number of ids with a component in each read file
    private static int count_ids() {
    	matched_keys = new ArrayList<String>();
		Object[] keys = first_ids.keySet().toArray();
		for (int i=0; i<keys.length; i++) {
			String new_key = (String) keys[i];
			new_key = new_key.substring(0,new_key.length()-1);
			new_key += "2";
			if (second_ids.containsKey(new_key)) matched_keys.add((String) keys[i]);
		}
		return matched_keys.size();
    }
    
    private static ArrayList<Integer> indexOf(ArrayList<Integer[]> a, Integer[] b) {
		ArrayList<Integer> retval = new ArrayList<Integer>();
		for (int i=0; i<a.size(); i++) {
			if (a.get(i).length == b.length) {
				boolean verify = true;
				for (int j=0; j<a.get(i).length; j++) {
					if (!a.get(i)[j].equals(b[j])) {
						verify = false;
					}
				}
				if (verify) {
					retval.add(i);
				}
			}
		}
		return retval;
    }
    
    public static void initialize_align(ArrayList<Integer[]> which, ArrayList<String> al) {
    	for (int i=0; i<al.size(); i++) {
    		which.add(new Integer[1]);
    	}
    }
	
	// Trim down the alignments to deal with the fasta_readlen issues
	private static void trim(ArrayList<String> al1, ArrayList<String> al2,
							 ArrayList<Integer> st1, ArrayList<Integer> st2, ArrayList<Integer> len1,
							 ArrayList<Integer> len2) {
		//HashMap<String,Integer> used_1 = new HashMap<String,Integer>();
		align1 = new ArrayList<Integer[]>(al1.size());
		align2 = new ArrayList<Integer[]>(al2.size());
		initialize_align(align1, al1);
		initialize_align(align2, al2);
		
		// first: trim alignments
		int x = 0;	// index into first/second_paths/start
		for (int i=0; i<matched_keys.size(); i++) {
			String id = matched_keys.get(i);
			ArrayList<Integer> index1 = first_ids.get(id);
			for (int j=0; j<index1.size(); j++) {
				// get path
				int index = index1.get(j).intValue();
				String[] match = al1.get(index).substring(1).split(":");
				Integer[] r = new Integer[match.length];
				for (int k=0; k<match.length; k++) {
					r[k] = Integer.decode(match[k]);
				}
				Integer[] t = null;
				t = Utility.adjust_alignment(r,st1.get(index),len1.get(index),s,fasta_readlen);
				align1.set(index,t);
					
				// get new start position
				int k=0;
				int new_st1 = st1.get(index);
				if (r[k].intValue() != t[0].intValue()) {
					new_st1 -= Math.min(s.getSegment(r[k].intValue()).length(),fasta_readlen-1);
					k++;
				}
				else {
					if (r.length == 1) {
						// we're fine, only aligned to one exon anyway
					}
					else {
						// it's an intron-bridging alignment, add the rest of the segment
						if (s.getSegment(r[k].intValue()).length()-(fasta_readlen-1) > 0) {
							new_st1 += s.getSegment(r[k].intValue()).length() - (fasta_readlen - 1);
						}
					}
				}
				while (r[k].intValue() != t[0].intValue() && new_st1 > 0) {
					new_st1 -= s.getSegment(r[k].intValue()).length();
					k++;
				}
				st1.set(index,new_st1);
			}
			
			id = id.substring(0,id.length()-1);
			id += "2";
			ArrayList<Integer> index2 = second_ids.get(id);
			for (int j=0; j<index2.size(); j++) {
				// get path
				int index = index2.get(j).intValue(); // this is not the READ index
				String[] match = al2.get(index).substring(1).split(":");
				Integer[] r = new Integer[match.length];
				for (int k=0; k<match.length; k++) {
					r[k] = Integer.decode(match[k]);
				}
				Integer[] t = null;
				t = Utility.adjust_alignment(r,st2.get(index),len2.get(index),s,fasta_readlen);
				align2.set(index,t);
				
				// get new start position
				int k=0;
				int new_st2 = st2.get(index);
				if (r[k].intValue() != t[0].intValue()) {
					new_st2 -= Math.min(s.getSegment(r[k].intValue()).length(),fasta_readlen-1);
					k++;
				}
				else {
					if (r.length == 1) {
						// we're fine, only aligned to one exon anyway
					}
					else {
						// it's an intron-bridging alignment, add the rest of the segment
						if (s.getSegment(r[k].intValue()).length()-(fasta_readlen-1) > 0) {
							new_st2 += s.getSegment(r[k].intValue()).length() - (fasta_readlen - 1);
						}
					}
				}
				while (r[k].intValue() != t[0].intValue() && new_st2 > 0) {
					new_st2 -= s.getSegment(r[k].intValue()).length();
					k++;
				}
				st2.set(index,new_st2);
			}
			x++;
		}
	}
	
	// third time's the charm
	//*
	private static void process_align(ArrayList<String> al1, ArrayList<String> al2,
									  ArrayList<Integer> st1, ArrayList<Integer> st2,
									  ArrayList<Integer> len1, ArrayList<Integer> len2) {
		// initialize everything we'll need for the alignments
		align = new ArrayList[reads.length];
		confidence = new ArrayList[reads.length];
		lengths = new ArrayList[reads.length];
		
		// first: trim alignments
		trim(al1,al2,st1,st2,len1,len2);
		
		// second: remove any duplicate alignments (side effect of fasta_readlen)
		ArrayList[] first_paths = new ArrayList[matched_keys.size()];
		ArrayList[] first_start = new ArrayList[matched_keys.size()];
		ArrayList[] first_conf = new ArrayList[matched_keys.size()];
		ArrayList[] second_paths = new ArrayList[matched_keys.size()];
		ArrayList[] second_start = new ArrayList[matched_keys.size()];
		ArrayList[] second_conf = new ArrayList[matched_keys.size()];
		
		for (int index=0; index<matched_keys.size(); index++) {
			String id = matched_keys.get(index);
			first_paths[index] = new ArrayList<Integer[]>();
			first_start[index] = new ArrayList<Integer>();
			first_conf[index] = new ArrayList<Double>();
			second_paths[index] = new ArrayList<Integer[]>();
			second_start[index] = new ArrayList<Integer>();
			second_conf[index] = new ArrayList<Double>();
			
			ArrayList<Integer> index_1 = first_ids.get(id);				// all records associated with this read id
			for (int j=0; j<index_1.size(); j++) {
				Integer[] path = align1.get(index_1.get(j).intValue());	// path associated with this record
				Integer start_pos = st1.get(index_1.get(j).intValue());	// start position associated with this record
				double quality = c1[index_1.get(j).intValue()];			// quality score associated with this record
				if (!Utility.contains(first_paths[index],path)) {
					// no problems, add everything
					first_paths[index].add(path);
					first_start[index].add(start_pos);
					first_conf[index].add(quality);
				}
				else {
					// this path exists; see if this is a different alignment location
					ArrayList<Integer> temp = indexOf(first_paths[index],path);
					boolean contained = false;
					for (int k=0; k<temp.size(); k++) {
						if (first_start[index].get(temp.get(k).intValue()).equals(start_pos)) {
							contained = true;
							break;
						}
					}
					if (!contained) {
						// different alignment to same path
						first_paths[index].add(path);
						first_start[index].add(start_pos);
						first_conf[index].add(quality);
					}
				}
			}
			
			id = id.substring(0,id.length()-1);
			id += "2";
			ArrayList<Integer> index_2 = second_ids.get(id);			// all records associated with this read id
			for (int j=0; j<index_2.size(); j++) {
				Integer[] path = align2.get(index_2.get(j).intValue());	// path associated with this record
				Integer start_pos = st2.get(index_2.get(j).intValue());	// start position associated with this record
				double quality = c2[index_2.get(j).intValue()];			// quality score associated with this record
				if (!Utility.contains(second_paths[index],path)) {
					// no problems, add everything
					second_paths[index].add(path);
					second_start[index].add(start_pos);
					second_conf[index].add(quality);
				}
				else {
					// this path exists; see if this is a different alignment location
					ArrayList<Integer> temp = indexOf(second_paths[index],path);
					boolean contained = false;
					for (int k=0; k<temp.size(); k++) {
						if (second_start[index].get(temp.get(k).intValue()).equals(start_pos)) {
							contained = true;
							break;
						}
					}
					if (!contained) {
						// different alignment to same path
						second_paths[index].add(path);
						second_start[index].add(start_pos);
						second_conf[index].add(quality);
					}
				}
			}
		}
		
		// third: find paths
		int index = 0;
		for (int i=0; i<first_paths.length; i++) {
			if (second_paths[i].size() > 0) {
				ArrayList[] paths = s.find_paths(first_paths[i],second_paths[i],first_conf[i],
												 second_conf[i],first_start[i],second_start[i],
												 reads[index][1].length());
				if (paths != null) {
					align[index] = paths[0];
					confidence[index] = paths[1];
					lengths[index] = paths[2];
					index++;
				}
			}
		}
	}
}