import re
import sys
import subprocess
import os
import itertools
import splicegraph
import gzip
import string
import table
import gff
import fasta
import fastq
import collections
import shutil

class Reference:
    def __init__(self, ref_name):
        try:
            index_filename = os.path.join(ref_name, "index.txt")
            self.genes = list(table.FileReader(index_filename))
        except IOError, e:
            raise Error("Could not read PSG reference named '%s'\n" % ref_name
                        + str(e))
        # Add internal gene_num field to genes
        for gene_num, gene in enumerate(self.genes, 1):
            gene["gene_num"] = gene_num

def learnpsg_classpath():
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), "LearnPSG")

def jvm_options():
    return ["-cp", learnpsg_classpath(),
            "-XX:+UseSerialGC"]

class Error(Exception):
    pass

# Returns the length of the first read in the given FASTQ file, or None if the file contains zero reads
def fastq_read_length(fastq_filename):
    with open(fastq_filename) as fastq_file:
        for read in fastq.Reader(fastq_file):
            return len(read.sequence.rstrip())
    return None

def reference_read_length_filename(ref_name):
    return os.path.join(ref_name, "max_read_length")

def write_reference_read_length(ref_name, max_read_length):
    with open(reference_read_length_filename(ref_name), 'w') as read_length_file:
        print >>read_length_file, max_read_length

def reference_read_length(ref_name):
    with open(reference_read_length_filename(ref_name)) as read_length_file:
        return int(read_length_file.read())

def alignments_by_gene(align_filename):
    read_id_field_num = 0
    id_field_num = 2
    stream = gzip.GzipFile(align_filename)
    curr_gene_num = None
    curr_alignments = []
    for line in stream:
        tokens = line.split('\t')
        global_sequence_id = tokens[id_field_num]
        gene_num, sequence_id = global_sequence_id.rsplit('.', 1)
        gene_num = int(gene_num)

        tokens[id_field_num] = sequence_id

        # Only use first word of read title line as id
        tokens[read_id_field_num] = first_word(tokens[read_id_field_num])
        
        if gene_num != curr_gene_num:
            if curr_alignments:
                yield curr_gene_num, curr_alignments
            curr_gene_num = gene_num
            curr_alignments = []

        curr_alignments.append(tokens)

    if curr_alignments:
        yield curr_gene_num, curr_alignments

    stream.close()

def first_word(s):
    return s.split()[0] if s else ""

def gene_alignments(genes, align_filename):
    alignments_iterator = alignments_by_gene(align_filename)
    next_gene, next_alignments = next(alignments_iterator, (None, None))
    for gene in genes:
        gene_num = gene["gene_num"]
        if gene_num != next_gene:
            yield []
        else:
            yield next_alignments
            next_gene, next_alignments = next(alignments_iterator, (None, None))
    if next_gene is not None:
        raise Error("Alignments were not sorted by gene")

def flatten(list_of_lists):
    result = []
    for lst in list_of_lists:
        result.extend(lst)
    return result
    
def gene_joined_alignments(genes, align_filenames):
    all_alignments = itertools.izip(*[gene_alignments(genes, align_filename)
                                      for align_filename in align_filenames])
    for alignments in all_alignments:
        yield flatten(alignments)
    
# def split_alignment(genes, align_filename, end, out_dir, gene_id_to_chrom_map,
#                     append=False):
#     for gene, alignments in itertools.izip(genes, gene_alignments(genes,
#                                                                   align_filename)):
#         gene_id = gene["gene_id"]
#         chrom = gene_id_to_chrom_map[gene_id]
#         chrom_dir = os.path.join(out_dir, chrom)
#         gene_filename = os.path.join(chrom_dir,
#                                      "%s.%d.align" % (gene_id, end))
#         gene_file = open(gene_filename, 'a' if append else 'w')
#         for alignment in sorted(alignments):
#             gene_file.write('\t'.join(alignment))
#         gene_file.close()

def parsed_em_stream(stream):
    def parse_field(line, pat, value_type):
        m = pat.match(line)
        if m is None:
            raise Error("Could not match line with pat:\n%s\n%s" % (line, pat))
        return value_type(m.group(1))
    
    lines = list(stream)
    d = {}
    fp = r'([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
    beta_pat = re.compile(r'beta = ' + fp)
    num_reads_pat = re.compile(r'\tNumber of reads: (\d+)')
    log_likelihood_pat = re.compile(r'Current log likelihood: ' + fp)
    iterations_pat = re.compile(r'Converged after (\d+) iterations')
    log_likelihood_diff_pat = re.compile(r'Log likelihood diff: ' + fp)
    max_edge_diff_pat = re.compile(r'Max edge difference: ' + fp)
    l1_norm_pat = re.compile(r'L1 norm: ' + fp)
    l2_norm_pat = re.compile(r'L2 norm: ' + fp)

    if lines[0].startswith("beta"): 
        d["beta"] = parse_field(lines[0], beta_pat, float)
        offset = 1
    else:
        offset = 0

    d["num_reads"] = parse_field(lines[offset], num_reads_pat, int)
    d["log_likelihood"] = parse_field(lines[-6], log_likelihood_pat, float)
    d["num_iterations"] = parse_field(lines[-5], iterations_pat, int)
    d["log_likelihood_diff"] = parse_field(lines[-4], log_likelihood_diff_pat, float)
    d["max_edge_diff"] = parse_field(lines[-3], max_edge_diff_pat, float)
    d["l1_norm"] = parse_field(lines[-2], l1_norm_pat, float)
    d["l2_norm"] = parse_field(lines[-1], l2_norm_pat, float)

    iterations = []
    for line in lines[offset + 1:-7]:
        tokens = line.split()
        log_likelihood = float(tokens[1])
        parameters = map(float, tokens[3:])
        iterations.append((log_likelihood, parameters))
    d["iterations"] = iterations
    return d
    
def parsed_em_output(filename):
    with open(filename) as f:
        return parsed_em_stream(f)

def infer_frequencies(genes,
                      alignments,
                      sample_name,
                      ref_name,
                      fragment_length_mean,
                      fragment_length_stddev,
                      beta,
                      pool=None,
                      tempdir=None):
    read_length = reference_read_length(ref_name)

    if tempdir is None:
        tempdir = sample_name
    
    gene_results_filename = os.path.join(sample_name, "gene.results.txt")
    gene_results_fields = ["gene_id",
                           "chrom",
                           "start",
                           "end",
                           "strand",
                           "num_transcripts",
                           "num_paths",
                           "num_parameters",
                           "num_reads",
                           "log_likelihood",
                           "parameters"]
    gene_results_writer = table.FileWriter(gene_results_filename,
                                           gene_results_fields)

    gene_results_debug_filename = os.path.join(sample_name, "gene.results.debug.txt")
    gene_results_debug_fields = (gene_results_fields +
                                 ["num_iterations",
                                  "log_likelihood_diff",
                                  "max_edge_diff",
                                  "l1_norm",
                                  "l2_norm"])
    gene_results_debug_writer = table.FileWriter(gene_results_debug_filename,
                                                 gene_results_debug_fields)


    isoform_results_filename = os.path.join(sample_name, "isoform.results.txt")
    isoform_results_fields = ["gene_id",
                              "transcript_id",
                              "prob"]
    isoform_results_writer = table.FileWriter(isoform_results_filename,
                                              isoform_results_fields)
    
    inferrer = PSGFrequencyInferrer(sample_name,
                                    ref_name,
                                    read_length,
                                    fragment_length_mean,
                                    fragment_length_stddev,
                                    beta,
                                    tempdir)

    gene_alignments = itertools.izip(genes, *alignments)
    
    if pool is not None:
        iterator = pool.imap_unordered(inferrer, gene_alignments, chunksize=10)
    else:
        iterator = itertools.imap(inferrer, gene_alignments)

    results = []
    for result in iterator:
        log(result["command"])
        if not result["success"]:
            log("Error: AlignPaired failed on gene %d with message: %s" %
                (result["gene_num"], result["error_message"]))
            continue
        if result["log"]:
            log("Warning: AlignPaired generated log info for gene %d: %s" %
                (result["gene_num"], result["log"]))
        gene_results_debug_writer.write(result)
        results.append(result)
    gene_results_debug_writer.close()

    # Order results by gene_num, as they may have been shuffled due to
    # the use of imap_unordered
    results.sort(key=lambda result: result["gene_num"])
    for result in results:
        gene_results_writer.write(result)
        for transcript_id, prob in result["isoform_probs"]:
            isoform_results_writer.write({"gene_id": result["gene_id"],
                                          "transcript_id": transcript_id,
                                          "prob": prob})
    gene_results_writer.close()
    isoform_results_writer.close()

def create_empty_file(filename):
    f = open(filename, 'w')
    f.close()

def log(s):
    print >>sys.stderr, s
        
def run_command(cmd, env=None):
    try:
        subprocess.check_call(cmd, shell=True, env=env)
    except subprocess.CalledProcessError, e:
        raise Error("Command was terminated or failed:\n%s" % cmd)

def write_alignments(filename, alignments):
    with open(filename, 'w') as alignment_file:
        for alignment in sorted(alignments):
            alignment_file.write('\t'.join(alignment))
    
def infer_gene_frequencies(gene,
                           alignments,
                           sample_name,
                           ref_name,
                           read_length,
                           fragment_length_mean,
                           fragment_length_stddev,
                           beta,
                           tempdir):
    # We can only handle single-end or paired-end data
    assert(1 <= len(alignments) <= 2)

    # Construct input and output filenames
    gene_num = gene["gene_num"]
    chrom = gene["chrom"]
    psg_src_filename = os.path.join(ref_name, chrom, "%d.psg" % gene_num)
    psg_filename = os.path.join(tempdir, "%d.psg" % gene_num)
    align_filenames = [os.path.join(tempdir, "%d.%d.align" % (gene_num, end))
                       for end in xrange(1, len(alignments) + 1)]
    em_filename = os.path.join(tempdir, "%d.em.out" % gene_num)
    em_log_filename = em_filename + ".log"

    # Construct Java program command
    if len(alignments) == 1: # single-end data
        arguments = ["Align",
                     psg_filename,
                     align_filenames[0],
                     read_length,
                     beta]
    else: # paired-end data
        arguments = ["AlignPaired",
                     psg_filename,
                     align_filenames[0],
                     align_filenames[1],
                     read_length,
                     fragment_length_mean,
                     fragment_length_stddev,
                     beta]
    argument_string = ' '.join(map(str, jvm_options() + arguments))
    cmd = "java %s > %s 2> %s" % (argument_string,
                                  em_filename,
                                  em_log_filename)
    
    result = gene.copy()
    result["command"] = cmd

    try:
        shutil.copy(psg_src_filename, psg_filename)
        for align_filename, end_alignments in zip(align_filenames, alignments):
            write_alignments(align_filename, end_alignments)
        run_command(cmd)
        
        with open(em_log_filename) as log_file:
            result["log"] = log_file.read()
        result.update(parsed_em_output(em_filename))
        result["success"] = True

        edge_probs = result["iterations"][-1][1]
        result["parameters"] = ",".join(["%.3g" % p for p in edge_probs])

        # Create PSG with learned parameters
        psg_file = open(psg_filename)
        psg = splicegraph.read_splice_graph(psg_file)
        psg_file.close()
        assert(len(psg.edges) == len(edge_probs))
        psg.setEdgeProbs(edge_probs)
        psg.normalize()

        # Write PSG file with learned parameters
        # sample_psg_filename = os.path.join(sample_name, chrom, gene_num + ".psg")
        # sample_psg_file = open(sample_psg_filename, 'w')
        # splicegraph.write_splice_graph(sample_psg_file, psg)
        # sample_psg_file.close()

        # Compute path probabilities of annotated transcripts
        isoform_probs = []
        psg_paths_filename = os.path.join(ref_name, chrom, "%d.paths" % gene_num)
        psg_paths_file = open(psg_paths_filename)
        for line in psg_paths_file:
            transcript_id, path_string = line.split('\t')
            path = map(int, path_string.split())
            path_prob = psg.subpathProb(path)
            isoform_probs.append((transcript_id, path_prob))
        psg_paths_file.close()
        result["isoform_probs"] = isoform_probs

        # Remove input and output files
        os.remove(psg_filename)
        for align_filename in align_filenames:
            os.remove(align_filename)
        os.remove(em_filename)
        os.remove(em_log_filename)
    except Exception, e:
        result["success"] = False
        result["error_message"] = str(e)
    
    return result

class PSGFrequencyInferrer:
    def __init__(self, sample_name, ref_name,
                 read_length, fragment_length_mean, fragment_length_stddev, beta,
                 tempdir):
        self.sample_name = sample_name
        self.ref_name = ref_name
        self.read_length = read_length
        self.fragment_length_mean = fragment_length_mean
        self.fragment_length_stddev = fragment_length_stddev
        self.beta = beta
        self.tempdir = tempdir
    def __call__(self, gene_and_alignments):
        return infer_gene_frequencies(gene_and_alignments[0],
                                      gene_and_alignments[1:],
                                      self.sample_name,
                                      self.ref_name,
                                      self.read_length,
                                      self.fragment_length_mean,
                                      self.fragment_length_stddev,
                                      self.beta,
                                      self.tempdir)

reverse_complement_translate_table = string.maketrans("ACGTNacgtn",
                                                      "TGCANtgcan")
def reverse_complement(s):
    return ''.join(reversed(s.translate(reverse_complement_translate_table)))

def chrom_sequence_from_dir(chrom, genome_dir):
    fasta_extensions = ["fa", "fasta"]
    for extension in fasta_extensions:
        filename = os.path.join(genome_dir, chrom + "." + extension)
        if os.path.exists(filename):
            fasta_file = open(filename)
            recs = list(fasta.Reader(fasta_file))
            if len(recs) != 1:
                raise Error("Chromosome FASTA file does not " +
                            "contain exactly one sequence: " +
                            filename)
            return recs[0].sequence
    else:
        raise Error((("Could not locate a FASTA file for chromosome '%s' " +
                      "in directory '%s'") % (chrom, genome_dir)))
        
class FastaDirectoryGenomeDB:
    def __init__(self, directory, cache_last_chrom=False):
        self.directory = directory
        self.cache_last_chrom = cache_last_chrom
        self.last_chrom = None
        self.last_chrom_sequence = None

    def sequence(self, chrom, start, end, strand):
        if self.last_chrom != chrom:
            chrom_sequence = chrom_sequence_from_dir(chrom, self.directory)
            if self.cache_last_chrom:
                self.last_chrom = chrom
                self.last_chrom_sequence = chrom_sequence
        else:
            chrom_sequence = self.last_chrom_sequence

        seq = chrom_sequence[start: end]
        return seq if strand == '+' else reverse_complement(seq)

def create_output_directory(dirname, dir_description=""):
    if os.path.exists(dirname):
        if dir_description:
            dir_description += " "
        raise Error("Could not create %soutput directory '%s' because it "
                    "already exists" % (dir_description, dirname))
    else:
        os.makedirs(dirname)
