#!/usr/bin/env python

import sys
import optparse
import collections
import os.path
import os
import subprocess
import multiprocessing
import itertools
import signal
import time
import tempfile

import gff
import splicegraph
import fasta
import string
import psg
import table

class WallClockTimer:
    def __init__(self):
        self.last_time = time.time()
    def elapsed_time(self):
        return time.time() - self.last_time
    def reset(self):
        old_last_time = self.last_time
        self.last_time = time.time()
        return self.last_time - old_last_time

def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

def run_bowtie(reads_filename, ref_name, align_filename, num_cores,
               temp_dir=None, sort_buffer_size=None):
    bowtie_prefix = os.path.join(ref_name, "references")
    cmd = ("bowtie --best --strata -p %d -l 25 -a -m 200 %s %s 2> %s" %
           (num_cores,
            bowtie_prefix,
            reads_filename,
            align_filename + ".log"))

    # Sort according to gene_num and only use tab delimiter because of
    # possible whitespace in read titles
    sort_cmd = "sort -k3,3n -t'\t'"

    if temp_dir is not None:
        sort_cmd += " -T " + temp_dir
    if sort_buffer_size is not None:
        sort_cmd += " -S " + sort_buffer_size
    cmd += " | " + sort_cmd
    cmd += " | gzip -c > " + align_filename
    psg.log(cmd)

    # Add LC_ALL=C to environment to force traditional sort order by "sort"
    env = os.environ.copy()
    env["LC_ALL"] = "C"
    
    psg.run_command(cmd, env)

def main():
    usage = "usage: %prog [options] reference_name sample_name mate1_read_file [mate2_read_file]"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("--frag-mean", dest="fragment_length_mean", 
                      help="Fragment length mean. "
                      "Used with paired-end data only. "
                      "(default: 200)",
                      type="int", default=200, metavar="MEAN")
    parser.add_option("--frag-stddev", dest="fragment_length_stddev",
                      help="Fragment length standard deviation. "
                      "Used with paired-end data only. "
                      "(default: 25)",
                      type="float", default=25.0, metavar="STDDEV")
    parser.add_option("--pseudocount", dest="beta",
                      help="Prior pseudocount for each edge.  Set this to "
                      "zero for maximum likelihood estimates.  Set to a "
                      "positive value to obtain smoother estimates when the "
                      "number of reads is small (default: 1)",
                      type="float", default=1, metavar="COUNT")
    parser.add_option("--tempdir", dest="tempdir",
                      help="Temporary directory for intermediate files "
                      "(default: %s)" % tempfile.gettempdir(),
                      type="str", default=tempfile.gettempdir(), metavar="DIR")
    parser.add_option("--sort-buffer-size", dest="sort_buffer_size",
                      help="Buffer size in memory for sorting alignments "
                      "(default: use default for 'sort' command)",
                      default=None, metavar="SIZE")
    parser.add_option("-p", "--num-threads", dest="num_threads",
                      help="Number of parallel threads to use "
                      "(default: %d)" % multiprocessing.cpu_count(),
                      default=multiprocessing.cpu_count(),
                      type=int, metavar="INT")

    options, args = parser.parse_args()
    if not (3 <= len(args) <= 4):
        parser.error("Invalid number of arguments")
    ref_name, sample_name = args[:2]
    read_filenames = args[2:]

    print >>sys.stderr, "Using %d cores" % options.num_threads
    pool = multiprocessing.Pool(options.num_threads, init_worker)

    try:
        timer = WallClockTimer()

        psg.create_output_directory(sample_name, "sample")

        process_name = "psginfer-%s-" % os.path.basename(sample_name)
        process_tempdir = tempfile.mkdtemp(prefix=process_name,
                                           dir=options.tempdir)

        # Read reference index
        reference_read_time = timer.reset()
        reference = psg.Reference(ref_name)
        psg.log("Reference reading time: %.0f seconds" % reference_read_time)

        # Run Bowtie
        align_filenames = []
        alignment_times = []
        for end_num, read_filename in enumerate(read_filenames, start=1):
            align_filename = os.path.join(sample_name,
                                          "alignment%d.bowtie.gz" % end_num)
            align_filenames.append(align_filename)
            run_bowtie(read_filename, ref_name, align_filename,
                       options.num_threads,
                       process_tempdir, options.sort_buffer_size)
            alignment_time = timer.reset()
            alignment_times.append(alignment_time)
            psg.log("Mate %d alignment time: %.0f seconds" % (end_num,
                                                              alignment_time))
        psg.log("Total alignment time: %.0f seconds" % sum(alignment_times))

        # Group alignments by gene (PSG)
        alignments = [psg.gene_alignments(reference.genes, align_filename)
                      for align_filename in align_filenames]

        # Run PSG inferrer
        psg.infer_frequencies(reference.genes,
                              alignments,
                              sample_name,
                              ref_name,
                              options.fragment_length_mean,
                              options.fragment_length_stddev,
                              options.beta,
                              pool,
                              process_tempdir)
        estimation_time = timer.reset()
        total_time = alignment_time + estimation_time + reference_read_time
        psg.log("Estimation time: %.0f seconds" % estimation_time)
        psg.log("Total time: %.0f seconds" % total_time)

        try:
            os.rmdir(process_tempdir)
        except OSError:
            pass
        
    except psg.Error, e:
        print >>sys.stderr, "Error: %s" % e.args[0]
        sys.exit(1)
    except KeyboardInterrupt:
        print >>sys.stderr, "Caught KeyboardInterrupt, terminating workers"
        pool.terminate()
        pool.join()
        sys.exit(1)

if __name__ == "__main__":
    main()
