#!/usr/bin/env python

import sys
import optparse
import collections
import os.path
import os
import subprocess
import multiprocessing
import itertools

import gff
import splicegraph
import fasta
import string
import table
import psg

def main():
    usage = "usage: %prog [options] reference_name"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("-a", "--annotations", dest="gtf_filename",
                      help="GTF file containing gene/isoform annotations. "
                      "(default: GTF is read from standard input)",
                      metavar="FILE")
    parser.add_option("-g", "--genome-dir", dest="genome_dir",
                      help="Directory containing chromosome FASTA files. "
                      "For each chromosome, CHROM, listed in the GTF file, "
                      "there must be a file CHROM.fa or CHROM.fasta in this "
                      "directory that contains the sequence of the "
                      "chromosome in FASTA format. "
                      "(default: .)",
                      default=".", metavar="DIR")
    parser.add_option("-k", dest="k",
                      help="Order of the PSGs to create.  "
                      "Line graphs correspond to k=0.  "
                      "Exon graphs correspond to k=1.  "
                      "For k>1, longer range dependencies are modeled. "
                      "Set k=-1 for full-length isoform graphs. "
                      "(default: 1)",
                      default=1, type=int, metavar="INT")
    parser.add_option("-u", "--unfactorized", dest="unfactorized",
                      help="Unfactorize the PSGs",
                      action="store_true")
    parser.add_option("-l", dest="max_read_length",
                      help="Maximum length of a read that will be aligned "
                      "against this reference (default: 100)",
                      default=100, type=int, metavar="INT")
    parser.add_option("-p", "--num-threads", dest="num_threads",
                      help="Number of parallel threads to use "
                      "(default: %d)" % multiprocessing.cpu_count(),
                      default=multiprocessing.cpu_count(),
                      type=int, metavar="INT")
    options, args = parser.parse_args()
    if len(args) != 1:
        parser.error("Invalid number of arguments")
    ref_name = args[0]

    print >>sys.stderr, "Using %d cores" % options.num_threads
    pool = multiprocessing.Pool(processes=options.num_threads)

    try:
        psg.create_output_directory(ref_name, "reference")

        # Record maximum read length
        psg.write_reference_read_length(ref_name, options.max_read_length)
        
        # Set up file to which all PSG reference sequences will be written
        fasta_filename = os.path.join(ref_name, "references.fasta")
        fasta_file = open(fasta_filename, 'w')
        fasta_writer = fasta.Writer(fasta_file)

        # Set up file for PSG index table
        index_filename = os.path.join(ref_name, "index.txt")
        index_fields = ["gene_id",
                        "chrom",
                        "start",
                        "end",
                        "strand",
                        "num_transcripts",
                        "num_paths",
                        "num_parameters"]
        index_writer = table.FileWriter(index_filename, index_fields)
        
        db = FastaDirectoryGenomeDB(options.genome_dir, cache_last_chrom=True)

        genes = genes_from_gtf(options.gtf_filename)

        # Sort genes by genomic coordinate and number according to order
        def gene_sort_key(gene):
            return (gene.chrom, gene.start(), gene.end(), gene.strand)
        sorted_genes = sorted(genes, key=gene_sort_key)
        gene_nums = dict([(g.gene_id, i)
                          for i, g in enumerate(sorted_genes, 1)])
        
        # Process genes by chromosome
        def gene_group_key(gene):
            return gene.chrom
        for chrom, chrom_genes in itertools.groupby(sorted_genes,
                                                    gene_group_key):
            psg.log("Processing chromosome %s..." % chrom)
            psg.log("\tCreating PSGs")
            create_psgs(pool, chrom, chrom_genes, db,
                        options.max_read_length, options.k,
                        options.unfactorized,
                        ref_name, fasta_writer, index_writer, gene_nums)
            
        fasta_file.close()
        index_writer.close()

        # Run bowtie-build on the FASTA file
        bowtie_prefix = os.path.join(ref_name, "references")
        cmd = "bowtie-build -q %s %s" % (fasta_filename, bowtie_prefix)
        psg.log(cmd)
        psg.run_command(cmd)
        
    except psg.Error, e:
        print >>sys.stderr, "Error: %s" % e.args[0]
        sys.exit(1)
    except KeyboardInterrupt:
        print >>sys.stderr, "Caught KeyboardInterrupt, terminating workers"
        pool.terminate()
        pool.join()
        sys.exit(1)

class Gene:
    def __init__(self, gene_id, chrom, strand, transcript_exons):
        self.gene_id = gene_id
        self.chrom = chrom
        self.strand = strand
        self.transcript_exons = transcript_exons

    def start(self):
        return min([exon_list[0][0] for exon_list in self.transcript_exons.values()])

    def end(self):
        return max([exon_list[-1][1] for exon_list in self.transcript_exons.values()])

    def num_transcripts(self):
        return len(self.transcript_exons)
    
class Transcript:
    def __init__(self, transcript_id, gene_id, chrom, strand, exons):
        self.transcript_id = transcript_id
        self.gene_id = gene_id
        self.chrom = chrom
        self.strand = strand
        self.exons = exons
        
exon_fields = ["gene_id", "transcript_id", "chrom", "start", "end", "strand"]
Exon = collections.namedtuple("Exon", exon_fields)

reverse_complement_translate_table = string.maketrans("ACGTNacgtn",
                                                      "TGCANtgcan")
def reverse_complement(s):
    return ''.join(reversed(s.translate(reverse_complement_translate_table)))

def chrom_sequence_from_dir(chrom, genome_dir):
    fasta_extensions = ["fa", "fasta"]
    for extension in fasta_extensions:
        filename = os.path.join(genome_dir, chrom + "." + extension)
        if os.path.exists(filename):
            fasta_file = open(filename)
            recs = list(fasta.Reader(fasta_file))
            if len(recs) != 1:
                raise psg.Error("Chromosome FASTA file does not " +
                                "contain exactly one sequence: " +
                                filename)
            return recs[0].sequence
    else:
        raise psg.Error((("Could not locate a FASTA file for chromosome '%s' " +
                          "in directory '%s'") % (chrom, genome_dir)))
        
class FastaDirectoryGenomeDB:
    def __init__(self, directory, cache_last_chrom=False):
        self.directory = directory
        self.cache_last_chrom = cache_last_chrom
        self.last_chrom = None
        self.last_chrom_sequence = None

    def sequence(self, chrom, start, end, strand):
        if self.last_chrom != chrom:
            chrom_sequence = chrom_sequence_from_dir(chrom, self.directory)
            if self.cache_last_chrom:
                self.last_chrom = chrom
                self.last_chrom_sequence = chrom_sequence
        else:
            chrom_sequence = self.last_chrom_sequence

        seq = chrom_sequence[start: end]
        return seq if strand == '+' else reverse_complement(seq)

def genes_from_gtf(filename):
    if filename is None:
        psg.log("Reading transcripts from GTF-formatted standard input...")
        return genes_from_gtf_stream(sys.stdin)
    else:
        psg.log("Reading transcripts from GTF-formatted file...")
        with open(filename) as stream:
            return genes_from_gtf_stream(stream)

def genes_from_gtf_stream(stream):
    exons = list(exons_from_gtf_stream(stream))
    transcripts = list(transcripts_from_exons(exons))
    genes = list(genes_from_transcripts(transcripts))
    psg.log("Read %d genes with %d transcripts and %d exons" %
            (len(genes), len(transcripts), len(exons)))
    return genes

def genes_from_transcripts(transcripts):
    def transcript_gene_id(transcript): return transcript.gene_id
    for gene_id, gene_transcripts_iter in sorted_and_grouped(transcripts, transcript_gene_id):
        try:
            gene_transcripts = list(gene_transcripts_iter)
            chrom = check_all_equal_attr(gene_transcripts, "chrom")
            strand = check_all_equal_attr(gene_transcripts, "strand")
            transcript_exons = dict([(t.transcript_id, t.exons)
                                     for t in gene_transcripts])
            yield Gene(gene_id, chrom, strand, transcript_exons)
        except UnequalAttributeError, e:
            raise psg.Error("Records for gene '%s' have multiple values "
                            "for attribute '%s': %s" % (gene_id,
                                                        e.attribute_name,
                                                        ', '.join(e.values)))
        
def transcripts_from_exons(exons):
    def exon_transcript_id(exon): return exon.transcript_id
    for transcript_id, transcript_exons_iter in sorted_and_grouped(exons, exon_transcript_id):
        try:
            transcript_exons = list(transcript_exons_iter)
            gene_id = check_all_equal_attr(transcript_exons, "gene_id")
            chrom = check_all_equal_attr(transcript_exons, "chrom")
            strand = check_all_equal_attr(transcript_exons, "strand")
            exon_intervals = sorted([(e.start, e.end) for e in transcript_exons])
            yield Transcript(transcript_id, gene_id, chrom, strand, exon_intervals)
        except UnequalAttributeError, e:
            raise psg.Error("Records for transcript '%s' have multiple values "
                            "for attribute '%s': %s" % (transcript_id,
                                                        e.attribute_name,
                                                        ', '.join(e.values)))

def exons_from_gtf_stream(stream):
    for rec in gff.Reader(stream):
        if rec.type == "exon":
            try:
                yield Exon(gene_id=rec.attributes["gene_id"][0],
                           transcript_id=rec.attributes["transcript_id"][0],
                           chrom=rec.seqid,
                           start=rec.start - 1, # 0-based coordinates internally
                           end=rec.end,
                           strand=rec.strand)
            except KeyError, e:
                raise MyError(("exon record is missing required " +
                               "attribute '%s'\n%s") % (e.args[0], repr(rec)))
        
def sorted_and_grouped(iterator, key):
    for group in itertools.groupby(sorted(iterator, key=key), key=key):
        yield group

class UnequalAttributeError(Exception):
    def __init__(self, attribute_name, values):
        self.attribute_name = attribute_name
        self.values = values
        
def check_all_equal_attr(records, attr_name):
    attr_values = list(set([getattr(rec, attr_name) for rec in records]))
    if len(attr_values) != 1:
        raise UnequalAttributeError(attr_name, sorted(attr_values))
    return attr_values[0]

class PSGCreator:
    def __init__(self, chrom, db, read_length, k, unfactorized, output_dir,
                 gene_nums):
        self.chrom = chrom
        self.db = db
        self.read_length = read_length
        self.output_dir = output_dir
        self.k = k
        self.unfactorized = unfactorized
        self.gene_nums = gene_nums
    def __call__(self, gene):
        gene_num = self.gene_nums[gene.gene_id]
        transcript_ids_and_exons = sorted(gene.transcript_exons.items())
        transcript_ids, exon_lists = zip(*transcript_ids_and_exons)
        
        log = ""

        # Create initial splice graph
        (gene_splice_graph, intervals, transcript_paths) = \
            splicegraph.kth_order_splice_graph(self.chrom,
                                               gene.strand,
                                               exon_lists,
                                               self.db,
                                               self.k)
        if self.unfactorized:
            if gene_splice_graph.getNumIsoforms() > 1000:
                log += "Warning: Did not unfactorize gene because it has more than 1000 paths: " + gene.gene_id + "\n"
            else:
                gene_splice_graph, transcript_paths = splicegraph.unfactored_splice_graph_and_paths(gene_splice_graph,
                                                                                                    transcript_paths)
        
        splice_graph_filename = os.path.join(self.output_dir,
                                             "%d.psg" % gene_num)
        splice_graph_file = open(splice_graph_filename, 'w')
        splicegraph.write_splice_graph(splice_graph_file, gene_splice_graph)
        splice_graph_file.close()

        transcript_paths_filename = os.path.join(self.output_dir,
                                                 "%d.paths" % gene_num)
        transcript_paths_file = open(transcript_paths_filename, 'w')
        for transcript_id, path in zip(transcript_ids, transcript_paths):
            line = "%s\t%s" % (transcript_id, ' '.join(map(str, path)))
            print >>transcript_paths_file, line
        transcript_paths_file.close()
        
        # Make FASTA file for this splice graph
        cmd = ("java %s MakeFastaStdOut %s %d" %
               (' '.join(psg.jvm_options()),
                splice_graph_filename,
                self.read_length))
        log += cmd + "\n"
        make_fasta_process = subprocess.Popen(cmd, shell=True,
                                              stdout=subprocess.PIPE)
        make_fasta_out = make_fasta_process.stdout
        raw_fasta_records = list(fasta.Reader(make_fasta_out))
        make_fasta_out.close()
        make_fasta_process.wait()
        if make_fasta_process.returncode != 0:
            raise psg.Error("Command was terminated or failed:\n%s" % cmd)

        raw_fasta_filename = os.path.join(self.output_dir, "%d.fasta" % gene_num)
        raw_fasta_file = open(raw_fasta_filename, 'w')
        raw_fasta_writer = fasta.Writer(raw_fasta_file)
        new_fasta_records = []
        for rec in raw_fasta_records:
            raw_fasta_writer.write(rec)
            new_title = "%d.%s" % (gene_num, rec.title)
            new_fasta_records.append(fasta.Record(new_title, rec.sequence))
        raw_fasta_file.close()

        return {"gene_num": gene_num,
                "gene_id": gene.gene_id,
                "chrom": self.chrom,
                "start": gene.start() + 1, # 1-based coordinates in output
                "end": gene.end(),
                "strand": gene.strand,
                "num_transcripts": gene.num_transcripts(),
                "num_paths": gene_splice_graph.getNumIsoforms(),
                "num_parameters": gene_splice_graph.getNumFreeParameters(),
                "fasta_records": new_fasta_records,
                "log": log}

def create_psgs(pool, chrom, genes, db, max_read_length, k, unfactorized,
                ref_name, fasta_writer, index_writer, gene_nums):
    chrom_output_dir = os.path.join(ref_name, chrom)
    if not os.path.exists(chrom_output_dir):
        os.makedirs(chrom_output_dir)
    creator = PSGCreator(chrom, db, max_read_length, k, unfactorized,
                         chrom_output_dir, gene_nums)
    for result in pool.imap(creator, genes, chunksize=10):
        sys.stderr.write(result["log"])
        for rec in result["fasta_records"]:
            fasta_writer.write(rec)
        index_writer.write(result)
    
if __name__ == "__main__":
    main()
